@@ -3,26 +3,54 @@ from typing import Union
 import math
 
 # TODO
-# Prefix conversion
-# MeasurementList error prop errors
-# MeasurementList type errors in regression
-# MeasurementList type compat 
-# ease of linreg with external
-# better looking plots
+# MeasurementList type compatability
 
 class Unit:
     "SI Unit taking in a string."
+    # Not Supported: mol (moles), cd (candela)
+    baseUnits = ['m','g','s','A','K']
+
+    derivedUnits = {
+    'J': ['kg m^2 s^-2'],
+    'V': ['kg m^2 s^-3 A^-1'],
+    'N': ['kg m s^-2'],
+    'W': ['kg m^2 s^-3'],
+    'T': ['kg s^-2 A^-1'],
+    'Pa': ['kg m^-1 s^-2'],
+    'Hz': ['s^-1'],
+    'C' : ['K'], # , lambda C: C + 273.15, lambda K: K - 273.15
+    'eV' : ['kg m^2 s^-2', 1.602176634e-19],
+    
+    }
+
+    knownUnits = list(derivedUnits.keys())
+    knownUnits += baseUnits
+
+    prefixes = {
+    # 'a':1e-18,
+    # 'f':1e-15,
+    'p':1e-12,
+    'n':1e-9,
+    'u':1e-6,
+    'm':1e-3,
+    'c':1e-2,
+    '' :1,
+    'k':1e3,
+    'M':1e6,
+    'G':1e9,
+    'T':1e12,
+    # 'P':1e15, # breaks parsing for Pa (Pascals)
+    # 'E':1e18
+    }
 
     def __init__(self,unitString: Union[str,dict]):
-        Unit.knownUnits = ['g','s','A','K','C','J','V','N','W','T','Pa','Hz','m']
-        Unit.prefixes = {'n':1e-9,'u':1e-6,'m':1e-3,'c':1e-2,'':1,'k':1e3,'M':1e6,'G':1e9}
-        
+
         # Given user string input, parse the units, prefixes and powers
         if(type(unitString) == str):
             self.units = dict.fromkeys(Unit.knownUnits)
             for unit in self.units:
-                self.units[unit] = {'prefix':'','power':0}
-            self.parse( unitString.replace(' ','').replace('{','').replace('}','') )
+                self.units[unit] = {'prefix':'', 'power':0}
+            self.parse(unitString.replace('{','(').replace('}',')'))
         
         # Used internally to construct a Unit from a dictionary of its units
         else:
@@ -42,53 +70,55 @@ class Unit:
     def parse(self,unitString):
         "Decompose string into its constituent SI units."
 
-        # Match a prefix
-        prefix = re.compile('([numckMG])')
+        # Match a prefix that is followed by a non-whitespace character
+        prefix = re.compile(f'([{"".join(prefix for prefix in Unit.prefixes.keys())}])\\B')
         # Match a known unit
-        unit = re.compile('([gsAKCJVNWTm]|(?:Pa)|(?:Hz))')
+        # Compiles to '([JVNWTmgsACK]|(?:Pa)|(?:Hz))' for default (base + derived) units
+        unit = re.compile(f"([{''.join([ unit_str if len(unit_str) == 1 else '' for unit_str in Unit.knownUnits])}]|{'|'.join([ ('(?:' + unit_str + ')') for unit_str in filter(lambda x: len(x) > 1, Unit.knownUnits) ])})") 
         # Match a '^' followed optionally by '-' and then any number of digits
-        rgxpower = re.compile('(\^)(\-?)(\d+)')
+        power = re.compile('(\^)(\-?)(\d+)')
+        # power = re.compile(r'\^(?:-?\d+)')
+        flip = False
 
         i = 0
         while i < len(unitString):
+            if (unitString[i] == '/'):
+                if (flip):
+                    raise Exception("labtex Unit Parsing Error: Cannot have two '/' characters.")
+                flip = True
+                i += 1
+
             prefixmatch = prefix.match(unitString[i:])
             prefixfound = prefixmatch is not None
+            # As all prefixes are of length 1
             unitmatch = unit.match(unitString[i+prefixfound:])
-            if (unitmatch is not None):
-                unitname = unitmatch.group(1)
-                powermatch = rgxpower.match(unitString[i+prefixfound+len(unitname):])
-                powerlength = powermatch.span()[1] if powermatch != None else 0
-                self.units[unitname] = {
-                    'prefix': prefixmatch.group(1) if prefixfound else '',
-                    'power': int(powermatch.group(2) + powermatch.group(3) if powermatch != None else 1)
-                    }
-                i += prefixfound + len(unitname) + powerlength
+            # print(f"unitmatch: {unitmatch}")
+            if (unitmatch is None):
+                if (unitString[i] in ['(',')']):
+                    raise Exception('labtex Unit Parsing Error: Parentheses are not supported. Use negative exponents or a \'/\' instead.')
+                if (unitString[i] not in [' ','1']):
+                    raise Exception(f'labtex Unit Parsing Error: Unknown character: \'{unitString[i]}\'. If this is intended, you can add it to the base units/prefixes with `Unit.baseUnits` and `Unit.prefixes`.')
+                i += 1
+                continue
+
+            prefixstr = prefixmatch.group(1) if prefixfound else ''
+            unitstr = unitmatch.group(1)
+            powermatch = power.match(unitString[i+prefixfound+len(unitstr):])
+
+            self.units[unitstr] = {
+                "prefix": prefixstr,
+                "power": (-1)**(2-flip) * int(powermatch.group(2) + powermatch.group(3) if powermatch else 1)
+            }
+
+            i += len(prefixstr) + len(unitstr) + (powermatch.span()[1] if powermatch else 0)
 
-            # account for 'm' as a prefix match but no succeeding unit
-            elif (prefixfound):
-                unitmatch = unit.match(unitString[i:])
-                if(unitmatch is not None):
-                    unitname = unitmatch.group(1)
-                    powermatch = rgxpower.match(unitString[i+len(unitname):])
-                    powerlength = powermatch.span()[1] if powermatch != None else 0
-                    self.units[unitname] = {
-                        'prefix': '',
-                        'power': int(powermatch.group(2) + powermatch.group(3) if powermatch else 1)
-                        }
-                    i += len(unitname) + powerlength
-                
-                else:
-                    raise Exception(f"Error in unit parsing with unknown characters: {unitString[i:]}")
-            else:
-                raise Exception(f"Error in unit parsing with unknown characters: {unitString[i:]}")
-        
     @staticmethod
     def unitless(self):
         return all([ dim['power'] == 0 for dim in self.units.values() ])
 
     @staticmethod
-    def singular(self): # only a single dimension has a non zero power
-        return sum([*map(lambda x: x['power'] != 0,self.units.values())]) == 1
+    def singular(self): # a unit with a single dimension
+        return sum([*map(lambda x: x['power'] != 0, self.units.values())]) == 1
 
     @staticmethod
     def singularunit(self):
@@ -111,7 +141,8 @@ class Unit:
         if(isinstance(obj,Unit)):
             newunits = {}
             for unit in Unit.knownUnits:
-                if((self.units[unit]["power"] > 0) ^ (obj.units[unit]["power"] > 0) ):
+                # If one of the units is unitless, return the other, where prefix='' and power=0 by default.
+                if((self.units[unit]["power"] != 0) ^ (obj.units[unit]["power"] != 0) ):
                     newunits[unit] = {
                         "prefix": self.units[unit]["prefix"] + obj.units[unit]["prefix"],
                         "power": self.units[unit]["power"] + obj.units[unit]["power"]
@@ -122,27 +153,10 @@ class Unit:
                         "power": self.units[unit]["power"] + obj.units[unit]["power"]
                         }
                 else:
-                    raise Exception("Measurements have different prefixes. Multiplication not supported.")
+                    raise Exception("Units have different prefixes. Multiplication not supported.")
             return Unit(newunits)
         else:
-            # We require only a single dimension present otherwise we dont know which dimension to change the prefix of.
-            if(Unit.singular(self) and int(math.log10(obj)) == math.log10(obj)):
-                singularunit = Unit.singularunit(self)
-                multiplicativefactor = obj**(1/singularunit["power"])
-                if(Unit.prefixes[singularunit["prefix"]] * multiplicativefactor in Unit.prefixes.values()):
-                    newunits = self.units.copy()
-                    for unit in Unit.knownUnits:
-                        if(self.units[unit] == singularunit):
-                            # find the key that corresponds to this value
-                            newunits[unit] = {
-                                "prefix": list(Unit.prefixes.keys())[list(Unit.prefixes.values()).index(
-                                Unit.prefixes[singularunit["prefix"]] * multiplicativefactor
-                            )],
-                                "power": self.units[unit]["power"]
-                            }
-
-                    return Unit(newunits)
-        return self
+            return self
 
     def __rmul__(self,obj):
         return self.__mul__(obj)
@@ -152,9 +166,9 @@ class Unit:
         if(isinstance(obj,Unit)):
             newunits = {}
             for unit in Unit.knownUnits:
-                if(self.units[unit]["power"] > 0 ^ obj.units[unit]["power"] > 0 ):
+                if((self.units[unit]["power"] != 0) ^ (obj.units[unit]["power"] != 0 )):
                     newunits[unit] = {
-                        "prefix": self.units[unit]["prefix"] + obj.units[unit]["power"],
+                        "prefix": self.units[unit]["prefix"] + obj.units[unit]["prefix"],
                         "power": self.units[unit]["power"] - obj.units[unit]["power"]
                     }
                 elif(self.units[unit]["prefix"] == obj.units[unit]["prefix"]):
@@ -188,3 +202,5 @@ class Unit:
         }
         return Unit(newunits)
 
+class U(Unit):
+    pass
\ No newline at end of file
